""" Ajax views for core_explore_keyword_app
"""
from django.contrib import messages
from django.http import JsonResponse
from django.shortcuts import render
from django.views import View
from mongoengine.base import NON_FIELD_ERRORS

from core_explore_keyword_app.components.search_operator import (
    api as search_operator_api,
)
from core_explore_keyword_app.components.search_operator.models import SearchOperator
from core_explore_keyword_app.views.admin.forms import SearchOperatorForm
from core_main_app.commons.exceptions import ApiError


class SearchOperatorConfigModalView(View):
    """View for saving the configuration of a search operator"""

    @staticmethod
    def post(request):
        """Method called after editing a search operator.

        Args:
            request:

        Returns:
        """
        operator_form = SearchOperatorForm(request.POST)

        if not operator_form.is_valid():
            return render(
                request,
                "core_explore_keyword_app/admin/search_ops_manager/modal/config/form.html",
                context={"form": operator_form},
            )
        else:
            action = "created"

            xpath_list = [
                xpath.strip()
                for xpath in operator_form.cleaned_data["xpath_list"].split("\n")
            ]

            operator = SearchOperator(
                name=operator_form.cleaned_data["name"], xpath_list=xpath_list
            )

            if "document_id" in request.POST and request.POST["document_id"] != "":
                operator.pk = operator_form.cleaned_data["document_id"]
                action = "edited"

            try:
                search_operator_api.upsert(operator)
            except ApiError as e:
                operator_form.add_error(NON_FIELD_ERRORS, str(e))
                return render(
                    request,
                    "core_explore_keyword_app/admin/search_ops_manager/modal/config/form.html",
                    context={"form": operator_form},
                )

            messages.add_message(
                request, messages.SUCCESS, "Operator successfully %s!" % action
            )

            return JsonResponse({})


class SearchOperatorDeleteModalView(View):
    """View for deleting a search operator"""

    @staticmethod
    def post(request):
        """Method called after confirmation of the deletion of a search operator.

        Args:
            request:

        Returns:
        """
        if "id" not in request.POST:
            messages.add_message(request, messages.ERROR, "Invalid delete request.")
            return JsonResponse({}, status=400)

        try:
            operator = search_operator_api.get_by_id(request.POST["id"])
            search_operator_api.delete(operator)

            messages.add_message(
                request, messages.INFO, "Operator %s deleted." % operator.name
            )
            return JsonResponse({}, status=200)
        except ApiError as e:
            messages.add_message(
                request, messages.ERROR, "Failed to delete operator: %s." % str(e)
            )
            return JsonResponse({}, status=500)



""" Fixtures files for Persistent Query Keyword
"""
from core_explore_keyword_app.components.persistent_query_keyword.models import (
    PersistentQueryKeyword,
)
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface


class PersistentQueryKeywordFixtures(FixtureInterface):
    """Persistent query keyword fixtures"""

    persistent_query_keyword_1 = None
    persistent_query_keyword_2 = None
    persistent_query_keyword_3 = None
    data_collection = None

    def insert_data(self):
        """Insert a set of Data.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_query_collection()

    def generate_query_collection(self):
        """Generate a Persistent Query Keyword collection.

        Returns:

        """

        # NOTE: no xml_content to avoid using unsupported GridFS mock
        self.persistent_query_keyword_1 = PersistentQueryKeyword(
            user_id="1", name="persistent_query_keyword_1"
        ).save()

        self.persistent_query_keyword_2 = PersistentQueryKeyword(
            user_id="2", name="persistent_query_keyword_2"
        ).save()

        self.persistent_query_keyword_3 = PersistentQueryKeyword(
            user_id="None", name="persistent_query_keyword_3"
        ).save()

        self.data_collection = [
            self.persistent_query_keyword_1,
            self.persistent_query_keyword_2,
            self.persistent_query_keyword_3,
        ]



""" Explore by Keyword forms
"""
from django import forms

from core_main_app.components.template_version_manager import (
    api as template_version_manager_api,
)


class KeywordForm(forms.Form):
    """
    Search by Keyword form
    """

    keywords = forms.CharField(widget=forms.TextInput(), required=False)
    query_id = forms.CharField(widget=forms.HiddenInput())
    user_id = forms.CharField(widget=forms.HiddenInput())
    global_templates = forms.MultipleChoiceField(
        widget=forms.CheckboxSelectMultiple(), required=False
    )
    user_templates = forms.MultipleChoiceField(
        widget=forms.CheckboxSelectMultiple(), required=False
    )
    order_by_field = forms.CharField(required=False, widget=forms.HiddenInput())

    def __init__(self, *args, **kwargs):
        """Init Keyword form"""
        request = kwargs.pop("request")
        super(KeywordForm, self).__init__(*args, **kwargs)

        # initialize template filters
        global_templates = [
            (template_version_manager.id, template_version_manager.title)
            for template_version_manager in template_version_manager_api.get_active_global_version_manager(
                request=request
            )
        ]
        user_templates = [
            (template_version_manager.id, template_version_manager.title)
            for template_version_manager in template_version_manager_api.get_active_version_manager_by_user_id(
                request=request
            )
        ]
        self.fields["global_templates"].choices = global_templates
        self.fields["user_templates"].choices = user_templates



""" Search Operator API
"""
from core_explore_keyword_app.components.search_operator.models import SearchOperator
from core_main_app.commons import exceptions
from core_main_app.utils.xml import xpath_to_dot_notation


def get_all():
    """Return all the Search Operator.

    Returns:
    """
    return SearchOperator.get_all()


def get_by_id(operator_id):
    """Return the Search Operator with the given id.

    Args:
        operator_id:

    Returns:
    """
    try:
        return SearchOperator.get_by_id(operator_id)
    except exceptions.ModelError:
        raise exceptions.ApiError("Invalid operator ID")
    except exceptions.DoesNotExist:
        raise exceptions.ApiError("Operator does not exist")


def get_by_name(operator_name):
    """Return the Search Operator with the given name.

    Args:
        operator_name:

    Returns:
    """
    try:
        return SearchOperator.get_by_name(operator_name)
    except exceptions.DoesNotExist:
        raise exceptions.ApiError("Operator does not exist")


def get_by_dot_notation_list(dot_notation_list):
    """Return the Search Operator with the given dot_notation list.

    Args:
        dot_notation_list:

    Returns:
    """
    try:
        return SearchOperator.get_by_dot_notation_list(dot_notation_list)
    except exceptions.DoesNotExist:
        raise exceptions.ApiError("Operator does not exist")


def upsert(operator):
    """Save or update a Search Operator.

    Args:
        operator:

    Returns:

    """
    try:
        # Create the dot notation list automatically from the XPath list
        operator.dot_notation_list = [
            xpath_to_dot_notation(xpath) for xpath in operator.xpath_list
        ]

        return operator.save_object()
    except exceptions.NotUniqueError:
        raise exceptions.ApiError("Operator name or xpath already exists")
    except exceptions.ModelError as model_error:
        raise exceptions.ApiError("Model save failed: %s" % str(model_error))


def delete(operator):
    """Delete a Search Operator.

    Args:
        operator:

    Returns:

    """
    return operator.delete()



""" Serializers used for the search operator REST API.
"""
from rest_framework_mongoengine.serializers import DocumentSerializer

from core_explore_keyword_app.components.search_operator import (
    api as search_operator_api,
)
from core_explore_keyword_app.components.search_operator.models import SearchOperator


class SearchOperatorSerializer(DocumentSerializer):
    """Search operator serializer"""

    class Meta(object):
        model = SearchOperator
        fields = ["id", "name", "xpath_list"]
        read_only_fields = ("id",)

    def create(self, validated_data):
        """Create and return a new `SearchOperator` instance, given the validated data."""
        # Create instance from the validated data and insert it in DB
        instance = SearchOperator(
            name=validated_data["name"], xpath_list=validated_data["xpath_list"]
        )
        search_operator_api.upsert(instance)

        return instance

    def update(self, instance, validated_data):
        """Update and return an existing `SearchOperator` instance, given the validated
        data.
        """
        instance.name = validated_data.get("name", instance.name)
        instance.xpath_list = validated_data.get("xpath_list", instance.xpath_list)
        return search_operator_api.upsert(instance)



""" Persistent Query Keyword model
"""
from mongoengine import errors as mongoengine_errors

from core_explore_common_app.components.abstract_persistent_query.models import (
    AbstractPersistentQuery,
)
from core_main_app.commons import exceptions


class PersistentQueryKeyword(AbstractPersistentQuery):
    """Persistent Query Keyword"""

    @staticmethod
    def get_by_id(query_id):
        """Get a persistent query Keyword

        Args:
            query_id:

        Returns:

        """
        try:
            return PersistentQueryKeyword.objects().get(pk=query_id)
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as e:
            raise exceptions.ModelError(str(e))

    @staticmethod
    def get_by_name(query_name):
        """Get a persistent query Keyword

        Args:
            query_name:

        Returns:

        """
        try:
            return PersistentQueryKeyword.objects().get(name=query_name)
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as e:
            raise exceptions.ModelError(str(e))

    @staticmethod
    def get_all():
        """Return all persistent query Keyword.

        Returns:

        """
        return PersistentQueryKeyword.objects.all()

    @staticmethod
    def get_all_by_user(user_id):
        """Return all persistent query Keyword by user.

        Args:
            user_id:

        Returns:

        """
        return PersistentQueryKeyword.objects(user_id=str(user_id))

    @staticmethod
    def get_none():
        """Return None object, used by data.

        Returns:

        """
        return PersistentQueryKeyword.objects().none()



from core_main_app.utils.databases.mongoengine_database import Database

SECRET_KEY = "fake-key"

INSTALLED_APPS = [
    # Django apps
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    # Local apps
    "tests",
]

CUSTOM_NAME = "Local"
""" :py:class:`str`: Name of the local instance
"""

# IN-MEMORY TEST DATABASE
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    },
}


MOCK_DATABASE_NAME = "db_mock"
MOCK_DATABASE_HOST = "mongomock://localhost"

database = Database(MOCK_DATABASE_HOST, MOCK_DATABASE_NAME)
database.connect()



""" Admin views for explore by keyword
"""
from django.views.generic import View

from core_explore_keyword_app.components.search_operator import (
    api as search_operator_api,
)
from core_explore_keyword_app.views.admin.forms import SearchOperatorForm
from core_main_app.utils.rendering import admin_render


class ListSearchOperatorsView(View):
    def get(self, request):
        """View to assign search operators to template

        Args:
            request:

        Returns:

        """
        context = {"title": "Search Operator Manager", "subtitle": "Operator List"}

        assets = {
            "js": [
                {
                    "path": "core_explore_keyword_app/admin/js/search_ops_manager.js",
                    "is_raw": False,
                },
                {
                    "path": "core_explore_keyword_app/admin/js/search_ops_manager_delete.js",
                    "is_raw": False,
                },
            ],
            "css": ["core_explore_keyword_app/admin/css/search_ops_manager.css"],
        }

        modals = [
            "core_explore_keyword_app/admin/search_ops_manager/modal/config/modal.html",
            "core_explore_keyword_app/admin/search_ops_manager/modal/delete/modal.html",
        ]

        try:
            context.update(
                {
                    "operators": search_operator_api.get_all(),
                    "operator_form": SearchOperatorForm(),
                }
            )

            return admin_render(
                request,
                "core_explore_keyword_app/admin/search_ops_manager.html",
                modals=modals,
                assets=assets,
                context=context,
            )
        except Exception as e:
            context.update({"error": "Unable to load the page: %s." % str(e).lower()})

            return admin_render(
                request,
                "core_main_app/admin/commons/errors/errors_wrapper.html",
                context=context,
            )



"""Explore keyword app Ajax views
"""
import json
import logging
import re

from django.urls import reverse_lazy
from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views.generic import View

import core_explore_keyword_app.permissions.rights as rights
import core_main_app.components.version_manager.api as version_manager_api
import core_main_app.utils.decorators as decorators
from core_explore_common_app.components.query import api as query_api
from core_explore_common_app.constants import LOCAL_QUERY_NAME
from core_explore_common_app.utils.query.query import send, create_local_data_source
from core_explore_common_app.views.user.ajax import CreatePersistentQueryUrlView
from core_explore_keyword_app.components.persistent_query_keyword.models import (
    PersistentQueryKeyword,
)
from core_explore_keyword_app.forms import KeywordForm
from core_main_app.components.template import api as template_api
from core_main_app.utils.databases.pymongo_database import get_full_text_query

logger = logging.getLogger("core_explore_keyword_app.views.user.ajax")


def _is_local_in_data_source(query):
    """Check if there is a data source that is local.

    Args:
        query:

    Returns:
    """
    # If we find a data source that is local
    for data_source in query.data_sources:
        # find local data source
        if data_source.name == LOCAL_QUERY_NAME:
            return True
    return False


def check_data_source(query):
    """Check the data sources. We will not provide suggestions if there are data sources selected but none of them is local.

    Args:
        query:

    Returns:
    """

    return len(query.data_sources) == 0 or _is_local_in_data_source(query)


class SuggestionsKeywordSearchView(View):
    @method_decorator(
        decorators.permission_required(
            content_type=rights.explore_keyword_content_type,
            permission=rights.explore_keyword_access,
            login_url=reverse_lazy("core_main_app_login"),
        )
    )
    def post(self, request, *args, **kwargs):
        """POST

        Args:
            request:
            *args:
            **kwargs:

        Returns:

        """
        suggestions = []
        search_form = KeywordForm(data=request.POST, request=request)
        keywords = request.POST.get("term")

        if search_form.is_valid():
            try:
                # get form values
                query_id = search_form.cleaned_data.get("query_id", None)
                global_templates = search_form.cleaned_data.get("global_templates", [])
                user_templates = search_form.cleaned_data.get("user_templates", [])

                # get all template version manager ids
                template_version_manager_ids = global_templates + user_templates

                # from ids, get all version manager
                version_manager_list = version_manager_api.get_by_id_list(
                    template_version_manager_ids, request=request
                )

                # from all version manager, build a list of all version (template)
                template_ids = []
                list([template_ids.extend(x.versions) for x in version_manager_list])

                if query_id is not None and keywords is not None:
                    # get query
                    query = query_api.get_by_id(query_id, request.user)

                    # Check the selected data sources
                    if check_data_source(query):

                        # Prepare query
                        query = self._get_query_prepared(
                            keywords, query, request, template_ids
                        )

                        # Send query
                        dict_results = send(
                            request, query, len(query.data_sources) - 1, 1
                        )

                        if dict_results["count"] > 0:
                            self._extract_suggestion_from_results(
                                dict_results, keywords, suggestions
                            )

            except Exception as e:
                logger.error("Exception while generating suggestions: " + str(e))

        return HttpResponse(
            json.dumps({"suggestions": suggestions}),
            content_type="application/javascript",
        )

    def _get_query_prepared(self, keywords, query, request, template_ids):
        """Prepare the query for suggestions.

        Args:
            keywords:
            query:
            request:
            template_ids:
        Returns:
        """

        # update query
        query.templates = template_api.get_all_accessible_by_id_list(
            template_ids, request=request
        )
        # TODO: improve query to get better results
        query.content = json.dumps(get_full_text_query(keywords))
        # Data source is local
        query.data_sources = [create_local_data_source(request)]
        return query

    def _extract_suggestion_from_results(self, dict_results, keywords, suggestions):
        """Extract suggestion from

        Args:
            dict_results:
            keywords:
            suggestions:

        Returns:
        """
        results = dict_results["results"]
        # Prepare keywords
        wordList = re.sub("[^\w]", " ", keywords).split()
        wordList = [x + "|" + x + "\w+" for x in wordList]
        wordList = "|".join(wordList)
        for result in results:
            # Extract suggestions from data
            listWholeKeywords = re.findall(
                "\\b(" + wordList + ")\\b", result["xml_content"], flags=re.IGNORECASE
            )
            labels = list(set(listWholeKeywords))

            for label in labels:
                label = label.lower()
                result_json = {}
                result_json["label"] = label
                result_json["value"] = label
                if not result_json in suggestions:
                    suggestions.append(result_json)


class CreatePersistentQueryUrlKeywordView(CreatePersistentQueryUrlView):
    """Create the persistent url from a Query"""

    view_to_reverse = "core_explore_keyword_results_redirect"

    @staticmethod
    def _create_persistent_query(query):
        # create the persistent query
        return PersistentQueryKeyword(
            user_id=query.user_id,
            content=query.content,
            templates=query.templates,
            data_sources=query.data_sources,
        )



""" Abstract Keyword Search Extras
"""
from abc import ABCMeta, abstractmethod


class AbstractKeywordSearchExtras(object, metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def get_extra_html():
        raise NotImplementedError("_get_extra_html method is not implemented.")

    @staticmethod
    @abstractmethod
    def get_extra_js():
        raise NotImplementedError("_get_extra_js method is not implemented.")

    @staticmethod
    @abstractmethod
    def get_extra_css():
        raise NotImplementedError("_get_extra_css method is not implemented.")



""" Search operators utilities
"""
import logging

from core_explore_keyword_app.components.search_operator import (
    api as search_operator_api,
)
from core_main_app.commons.exceptions import ApiError

LOGGER = logging.getLogger(__name__)


def build_search_operator_query(search_operator_name, value):
    search_operator = search_operator_api.get_by_name(search_operator_name)
    search_operator_list = [
        {search_operator_dot_notation: value}
        for search_operator_dot_notation in search_operator.dot_notation_list
    ]

    # Add `#text` field in case it is defined
    search_operator_list += [
        {"%s.#text" % search_operator_dot_notation: value}
        for search_operator_dot_notation in search_operator.dot_notation_list
    ]

    return {"$or": search_operator_list}


def get_keywords_from_search_operator_query(query):
    # Any query without $or does not contain keywords (since keyword and keyword.#text
    # are always part of the query, separated by $or).
    if "$or" not in query.keys():
        return None
    else:
        # The attribute of the $or sub_query must be a string : the search operator path
        for sub_query in query["$or"]:
            for key, value in sub_query.items():
                if not isinstance(sub_query[key], str):
                    return None

    dot_notation_list = [list(query_item.keys())[0] for query_item in query["$or"]]
    dot_notation_list = [
        dot_notation_item
        for dot_notation_item in dot_notation_list
        if not dot_notation_item.endswith("#text")
    ]
    value = list(query["$or"][0].values())[0]

    try:
        return "%s:%s" % (
            search_operator_api.get_by_dot_notation_list(dot_notation_list).name,
            value,
        )
    except ApiError as api_error:
        LOGGER.info("API error for query: %s (%s)" % (str(query), str(api_error)))
        return None



"""Template tag to include the search bar
"""

from django import template

from core_explore_common_app.components.query import api as query_api
from core_explore_common_app.components.query.models import Query
from core_explore_common_app.views.user.ajax import add_local_data_source
from core_explore_keyword_app.forms import KeywordForm
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.settings import DATA_SORTING_FIELDS

register = template.Library()


@register.inclusion_tag(
    "core_explore_keyword_app/user/embedded_search_bar.html", takes_context=True
)
def show_search_bar(context):
    """Include the search bar in a template.

    Args:
        context: Context

    Returns:

    """
    request = context["request"]
    try:
        # create Query
        query = Query(user_id=str(request.user.id), templates=[])

        # add local data source to the query
        add_local_data_source(request, query)

        # set visibility
        query_api.set_visibility_to_query(query, request.user)

        # upsert the query
        query_api.upsert(query, request.user)

        # create keyword form
        data_form = {
            "query_id": str(query.id),
            "user_id": query.user_id,
            "order_by_field": ",".join(DATA_SORTING_FIELDS),
        }
        search_form = KeywordForm(data=data_form, request=request)

        context = {
            "data": {
                "search_form": search_form,
                "query_id": search_form.data["query_id"],
            }
        }
    except AccessControlError as ace:
        context = {"data": {"error": str(ace)}}
    except:
        context = {"data": {"error": "An unexpected error occurred"}}

    return template.RequestContext(request, context)



"""Explore Keyword models
"""

from django.db import models

from core_explore_keyword_app.permissions import rights
from core_main_app.permissions.utils import get_formatted_name


class ExploreKeyword(models.Model):
    class Meta(object):
        verbose_name = "core_explore_keyword_app"
        default_permissions = ()
        permissions = (
            (
                rights.explore_keyword_access,
                get_formatted_name(rights.explore_keyword_access),
            ),
        )



""" Admin forms for Search Operators.
"""
from django import forms
from django_mongoengine.forms import DocumentForm
from core_main_app.utils.xml import validate_xpath
from core_main_app.commons import exceptions as core_main_app_exceptions

from core_explore_keyword_app.components.search_operator.models import SearchOperator


class SearchOperatorForm(DocumentForm):
    """Main form to edit search operators."""

    document_id = forms.CharField(required=False, widget=forms.HiddenInput())
    xpath_list = forms.CharField(
        help_text="Enter one xpath per line.",
        required=True,
        widget=forms.Textarea(
            attrs={"placeholder": "/path/to/xpath", "class": "form-control"}
        ),
    )

    class Meta(object):
        document = SearchOperator
        fields = ["name"]

        widgets = {
            "name": forms.TextInput(
                attrs={"placeholder": "Operator name", "class": "form-control"}
            )
        }

    def clean_xpath_list(self):
        """Validate the list of xpath.

        Returns:
        """
        data = self.cleaned_data["xpath_list"].split("\n")
        line_xpath = 0

        for xpath in data:
            line_xpath += 1
            try:
                validate_xpath(xpath)
            except core_main_app_exceptions.XMLError as e:
                raise forms.ValidationError(
                    "XPath syntax error (line %d): %s" % (line_xpath, str(e))
                )

        return self.cleaned_data["xpath_list"]



""" Search Operator model
"""
import re
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors
from core_main_app.commons import exceptions
from core_main_app.utils.xml import validate_xpath


def validate_xpath_list(xpath_list):
    item_position = 0

    for xpath in xpath_list:
        try:
            validate_xpath(xpath)
            item_position += 1
        except exceptions.XMLError as e:
            raise mongoengine_errors.ValidationError(
                "XPath syntax error (item #%d): %s" % (item_position, str(e))
            )


def validate_name(name):
    if re.match(r"^[a-zA-Z][a-zA-Z0-9]+$", name) is None:
        raise mongoengine_errors.ValidationError(
            "Name should only contains alpha-numerical characters."
        )


class SearchOperator(Document):
    """Search Operator model"""

    name = fields.StringField(blank=False, unique=True, validation=validate_name)
    xpath_list = fields.ListField(
        blank=False, unique=True, validation=validate_xpath_list
    )
    dot_notation_list = fields.ListField(blank=False, unique=True)

    @staticmethod
    def get_all():
        """Retrieve all search operators.

        Returns:
        """
        return SearchOperator.objects().all()

    @staticmethod
    def get_by_id(operator_id):
        """Retrieve a search operator with a given id.

        Args:
            operator_id:

        Returns:
        """
        try:
            return SearchOperator.objects.get(pk=operator_id)
        except mongoengine_errors.ValidationError as validation_error:
            raise exceptions.ModelError(str(validation_error))
        except mongoengine_errors.DoesNotExist as does_not_exist:
            raise exceptions.DoesNotExist(str(does_not_exist))

    @staticmethod
    def get_by_name(operator_name):
        """Retrieve all search operators with a given name.

        Args:
            operator_name:

        Returns:
        """
        try:
            return SearchOperator.objects.get(name=operator_name)
        except mongoengine_errors.DoesNotExist as does_not_exist:
            raise exceptions.DoesNotExist(str(does_not_exist))

    @staticmethod
    def get_by_dot_notation_list(operator_dot_notation_list):
        """Retrieve all search operators with a given dot_notation list.

        Args:
            operator_dot_notation_list:

        Returns:
        """
        try:
            return SearchOperator.objects.get(
                dot_notation_list=operator_dot_notation_list
            )
        except mongoengine_errors.DoesNotExist as does_not_exist:
            raise exceptions.DoesNotExist(str(does_not_exist))

    def save_object(self):
        """Upsert a search operator and handle possible issues.

        Returns:
            SearchOperator
        """
        try:
            return self.save()
        except mongoengine_errors.NotUniqueError as not_unique_error:
            raise exceptions.NotUniqueError(str(not_unique_error))
        except Exception as exception:
            raise exceptions.ModelError(str(exception))



"""Core Explore Keyword App views
"""
import json
from typing import Dict, Any, List

from django.http import HttpResponseRedirect
from django.urls import reverse_lazy, reverse
from django.utils.decorators import method_decorator

import core_explore_keyword_app.components.persistent_query_keyword.api as persistent_query_keyword_api
import core_explore_keyword_app.permissions.rights as rights
import core_main_app.components.version_manager.api as version_manager_api
import core_main_app.utils.decorators as decorators
from core_explore_common_app.components.query import api as query_api
from core_explore_common_app.settings import DEFAULT_DATE_TOGGLE_VALUE
from core_explore_common_app.utils.query.query import create_default_query
from core_explore_common_app.views.user.views import (
    ResultQueryRedirectView,
    ResultsView,
)
from core_explore_keyword_app.components.persistent_query_keyword.models import (
    PersistentQueryKeyword,
)
from core_explore_keyword_app.forms import KeywordForm
from core_explore_keyword_app.settings import EXPLORE_KEYWORD_APP_EXTRAS
from core_explore_keyword_app.utils.search_operators import (
    build_search_operator_query,
    get_keywords_from_search_operator_query,
)
from core_main_app.commons.exceptions import DoesNotExist, ApiError
from core_main_app.components.template import api as template_api
from core_main_app.settings import DATA_SORTING_FIELDS
from core_main_app.utils.rendering import render


class KeywordSearchView(ResultsView):
    query_builder_interface = "core_explore_keyword_app/user/search_bar.html"

    @method_decorator(
        decorators.permission_required(
            content_type=rights.explore_keyword_content_type,
            permission=rights.explore_keyword_access,
            login_url=reverse_lazy("core_main_app_login"),
        )
    )
    def get(self, request, **kwargs):
        """GET

        Args:
            request:
            **kwargs:

        Returns:

        """
        query_id = str(kwargs["query_id"]) if "query_id" in kwargs else None

        # assets / modals / forms
        context = self._get(request, query_id)

        return render(
            request,
            "core_explore_keyword_app/user/index.html",
            assets=self.assets,
            modals=self.modals,
            context=context,
        )

    @method_decorator(
        decorators.permission_required(
            content_type=rights.explore_keyword_content_type,
            permission=rights.explore_keyword_access,
            login_url=reverse_lazy("core_main_app_login"),
        )
    )
    def post(self, request):
        """POST

        Args:
            request:

        Returns:

        """

        # assets / modals / forms
        context = self._post(request)

        # if any errors or warning detected
        if context["error"] or context["warning"]:
            # render form again
            return render(
                request,
                "core_explore_keyword_app/user/index.html",
                assets=self.assets,
                modals=self.modals,
                context=context,
            )
        # no errors, redirect to search form
        return HttpResponseRedirect(
            reverse(
                "core_explore_keyword_app_search",
                kwargs={"query_id": context["query_id"]},
            )
        )

    @staticmethod
    def _parse_query(query_content):
        keyword_list = list()
        query_json = json.loads(query_content)

        if "$text" in query_json:
            keywords = query_json["$text"]["$search"]
            keyword_list = [
                "{}".format(keyword)
                for keyword in keywords.split('"')
                if keyword not in ("", " ")
            ]

        if "$and" in query_json:
            result = []

            # parse all the sub_queries
            for sub_query in query_json["$and"]:
                result.append(KeywordSearchView._parse_query(json.dumps(sub_query)))

            return ",".join(result)

        elif len(query_json.keys()) != 0:  # Avoid parsing empty query
            keyword = get_keywords_from_search_operator_query(query_json)

            if keyword is not None:
                keyword_list.append(keyword)

        return ",".join(keyword_list)

    def _get(self, request, query_id):
        """Prepare the GET context

        Args:
            query_id:

        Returns:

        """
        try:
            error = None
            # set the correct default ordering for the context
            default_order = ",".join(DATA_SORTING_FIELDS)
            if query_id is None:
                # create query
                query = create_default_query(request, [])
                # upsert the query
                query_api.upsert(query, request.user)
                # create keyword form
                # create all data for select values in forms
                keywords_data_form = {
                    "query_id": str(query.id),
                    "user_id": query.user_id,
                }
            else:  # query_id is not None
                # get the query id
                query = query_api.get_by_id(query_id, request.user)
                user_id = query.user_id

                # get all keywords back
                keywords = self._parse_query(query.content)

                # get all version managers
                version_managers = []
                for template in query.templates:
                    version_managers.append(
                        str(
                            version_manager_api.get_from_version(
                                template, request=request
                            ).id
                        )
                    )
                # create all data for select values in forms
                keywords_data_form = {
                    "query_id": str(query.id),
                    "user_id": user_id,
                    "keywords": keywords,
                    "global_templates": version_managers,
                    "order_by_field": super().build_sorting_context_array(query),
                    "user_templates": version_managers,
                }
                # set the correct ordering for the context
                if keywords_data_form["order_by_field"] != 0:
                    default_order = keywords_data_form["order_by_field"]

        except Exception as e:
            error = "An unexpected error occurred while loading the query: {}.".format(
                str(e)
            )
            return {"error": error}

        search_form = KeywordForm(data=keywords_data_form, request=request)
        return self._format_keyword_search_context(
            search_form, error, None, default_order
        )

    def _post(self, request):
        """Prepare the POST context

        Args:
            request:

        Returns:

        """
        error = None
        warning = None
        search_form = KeywordForm(data=request.POST, request=request)
        # validate form
        if search_form.is_valid():
            try:
                # get form values
                query_id = search_form.cleaned_data.get("query_id", None)
                keywords = search_form.cleaned_data.get("keywords", None)
                global_templates = search_form.cleaned_data.get("global_templates", [])
                user_templates = search_form.cleaned_data.get("user_templates", [])
                order_by_field_array = (
                    search_form.cleaned_data.get("order_by_field", "")
                    .strip()
                    .split(";")
                )
                # get all template version manager ids
                template_version_manager_ids = global_templates + user_templates
                # from ids, get all version manager
                version_manager_list = version_manager_api.get_by_id_list(
                    template_version_manager_ids, request=request
                )
                # from all version manager, build a list of all version (template)
                template_ids = []
                list([template_ids.extend(x.versions) for x in version_manager_list])
                if query_id is None or keywords is None:
                    error = "Expected parameters are not provided"
                else:
                    # get query
                    query = query_api.get_by_id(query_id, request.user)
                    if len(query.data_sources) == 0:
                        warning = "Please select at least 1 data source."
                    else:
                        # update query
                        query.templates = template_api.get_all_accessible_by_id_list(
                            template_ids, request=request
                        )
                        keywords_list = keywords.split(",") if keywords else []
                        query.content = self._build_query(keywords_list)
                        # set the data-sources filter value according to the POST request field
                        for data_sources_index in range(len(query.data_sources)):
                            # update the data-source filter only if it's not a new data-source
                            # (the default filter value is already added when the data-source
                            # is created)
                            if data_sources_index in range(
                                0, len(order_by_field_array)
                            ):
                                query.data_sources[
                                    data_sources_index
                                ].order_by_field = order_by_field_array[
                                    data_sources_index
                                ]

                        query_api.upsert(query, request.user)
            except DoesNotExist:
                error = "An unexpected error occurred while retrieving the query."
            except Exception as e:
                error = "An unexpected error occurred: {}.".format(str(e))
        else:
            error = "An unexpected error occurred: the form is not valid."

        return self._format_keyword_search_context(
            search_form,
            error,
            warning,
            search_form.cleaned_data.get("order_by_field", "").strip(),
        )

    @staticmethod
    def _build_query(initial_keyword_list):
        """Build query content with correct keywords and search operators

        Args:
            initial_keyword_list:

        Returns:
        """
        main_query = list()
        keyword_list = list()

        for keyword in initial_keyword_list:
            if ":" in keyword:
                try:
                    split_keyword = keyword.split(":")

                    main_query.append(
                        build_search_operator_query(split_keyword[0], split_keyword[1])
                    )
                except ApiError:
                    keyword_list.append(keyword)
            else:
                keyword_list.append(keyword)

        keyword_query = ['"' + keyword + '"' for keyword in keyword_list]

        if len(keyword_query) > 0:
            keyword_query = {"$text": {"$search": " ".join(keyword_query)}}
            main_query.append(keyword_query)

        # If the query is empty, match all documents
        if len(main_query) == 0:
            return json.dumps({})
        elif len(main_query) == 1:  # If there is one query item, match one this item.
            return json.dumps(main_query[0])
        else:  # For multiple items, a "$and" query is needed.
            return json.dumps({"$and": main_query})

    def _load_assets(self):
        """Return assets structure

        Returns:

        """
        assets = super()._load_assets()
        extra_assets: Dict[str, List[Any]] = {
            "js": [
                {
                    "path": "core_explore_keyword_app/libs/tag-it/2.0/js/tag-it.js",
                    "is_raw": True,
                },
                {
                    "path": "core_explore_keyword_app/libs/stretchy/stretchy.min.js",
                    "is_raw": False,
                },
                {
                    "path": "core_explore_keyword_app/user/js/search/search.js",
                    "is_raw": False,
                },
                {
                    "path": "core_explore_keyword_app/user/js/search/autocomplete_source.js",
                    "is_raw": False,
                },
                {
                    "path": "core_explore_keyword_app/user/js/persistent_query.raw.js",
                    "is_raw": True,
                },
            ],
            "css": [
                "core_explore_keyword_app/libs/tag-it/2.0/css/jquery.tagit.css",
                "core_explore_keyword_app/user/css/search/search.css",
            ],
        }

        assets["js"].extend(extra_assets["js"])
        assets["css"].extend(extra_assets["css"])
        for extra_app in EXPLORE_KEYWORD_APP_EXTRAS:
            assets["js"].extend(extra_app.get_extra_js())
            assets["css"].extend(extra_app.get_extra_css())

        return assets

    def _load_modals(self):
        """Return modals structure

        Returns:

        """
        return super()._load_modals()

    def _format_keyword_search_context(
        self, search_form, error, warning, query_order=""
    ):
        """Format the context for the keyword research page

        Args:
            search_form:
            error:
            warning:
            query_order:

        Returns:

        """
        context = {
            "search_form": search_form,
            "query_id": search_form.data["query_id"],
            "error": error,
            "warning": warning,
            "default_date_toggle_value": DEFAULT_DATE_TOGGLE_VALUE,
            "data_sources_selector_template": "core_explore_common_app/user/selector/data_sources_selector.html",
            "data_sorting_fields": query_order,
            "default_data_sorting_fields": ",".join(DATA_SORTING_FIELDS),
            "query_builder_interface": self.query_builder_interface,
            "EXPLORE_KEYWORD_APP_EXTRAS_HTML": [
                path
                for extra in EXPLORE_KEYWORD_APP_EXTRAS
                for path in extra.get_extra_html()
            ],
        }

        return context


class ResultQueryRedirectKeywordView(ResultQueryRedirectView):
    model_name = PersistentQueryKeyword.__name__
    object_name = "persistent_query_keyword"
    redirect_url = "core_explore_keyword_app_search"

    @method_decorator(
        decorators.permission_required(
            content_type=rights.explore_keyword_content_type,
            permission=rights.explore_keyword_access,
            login_url=reverse_lazy("core_main_app_login"),
        )
    )
    def get(self, request, *args, **kwargs):
        return super(ResultQueryRedirectKeywordView, self).get(
            self, request, *args, **kwargs
        )

    @staticmethod
    def _get_persistent_query_by_id(persistent_query_id, user):
        return persistent_query_keyword_api.get_by_id(persistent_query_id, user)

    @staticmethod
    def _get_persistent_query_by_name(persistent_query_name, user):
        return persistent_query_keyword_api.get_by_name(persistent_query_name, user)

    @staticmethod
    def get_url_path():
        return reverse(
            ResultQueryRedirectKeywordView.redirect_url, kwargs={"query_id": "query_id"}
        ).split("query_id")[0]

    @staticmethod
    def _get_reversed_url(query):
        return reverse(
            ResultQueryRedirectKeywordView.redirect_url, kwargs={"query_id": query.id}
        )

    @staticmethod
    def _get_reversed_url_if_failed():
        return reverse("core_explore_keyword_app_search")



from unittest import TestCase
from unittest.mock import patch, Mock
from core_explore_keyword_app.components.search_operator import (
    api as search_operator_api,
)
from core_explore_keyword_app.components.search_operator.models import SearchOperator
from core_explore_keyword_app.utils.search_operators import (
    build_search_operator_query,
    get_keywords_from_search_operator_query,
)
from core_main_app.commons.exceptions import ApiError


class TestBuildSearchOperatorQuery(TestCase):
    @patch.object(search_operator_api, "get_by_name")
    def test_single_dot_notation_returns_valid_query(self, mock_get_by_name):
        mock_dot_notation = "mock.dot.notation"
        mock_value = "mock_value"
        mock_search_operator = Mock(spec=SearchOperator)
        mock_search_operator.dot_notation_list = [mock_dot_notation]
        mock_get_by_name.return_value = mock_search_operator

        expected_dict = {
            "$or": [
                {mock_dot_notation: mock_value},
                {"%s.#text" % mock_dot_notation: mock_value},
            ]
        }

        returned_dict = build_search_operator_query("mock_name", mock_value)

        self.assertDictEqual(returned_dict, expected_dict)

    @patch.object(search_operator_api, "get_by_name")
    def test_multi_dot_notation_returns_valid_query(self, mock_get_by_name):
        mock_dot_notation_list = ["mock.dot.notation.1", "mock.dot.notation.2"]
        mock_value = "mock_value"
        mock_search_operator = Mock(spec=SearchOperator)
        mock_search_operator.dot_notation_list = mock_dot_notation_list
        mock_get_by_name.return_value = mock_search_operator

        mock_dot_notation_values = [
            {mock_dot_notation: mock_value}
            for mock_dot_notation in mock_dot_notation_list
        ]
        mock_dot_notation_values += [
            {"%s.#text" % mock_dot_notation: mock_value}
            for mock_dot_notation in mock_dot_notation_list
        ]

        expected_dict = {"$or": mock_dot_notation_values}

        returned_dict = build_search_operator_query("mock_name", mock_value)

        self.assertDictEqual(returned_dict, expected_dict)


class TestGetKeywordsFromSearchOperatorQuery(TestCase):
    def test_query_without_or_returns_none(self):
        self.assertEqual(get_keywords_from_search_operator_query({}), None)

    @patch.object(search_operator_api, "get_by_dot_notation_list")
    def test_returns_valid_keyword(self, mock_get_by_dot_notation_list):
        mock_keyword = "mock_keyword"
        mock_search_operator = Mock(spec=SearchOperator)
        mock_search_operator.name = mock_keyword
        mock_get_by_dot_notation_list.return_value = mock_search_operator
        mock_value = "mock_value"
        mock_query = {
            "$or": [
                {"mock.dot.notation.1": mock_value},
                {"mock.dot.notation.1.#text": mock_value},
                {"mock.dot.notation.2": mock_value},
                {"mock.dot.notation.2.#text": mock_value},
            ]
        }

        expected_string = "%s:%s" % (mock_keyword, mock_value)
        returned_string = get_keywords_from_search_operator_query(mock_query)

        self.assertEqual(returned_string, expected_string)

    @patch.object(search_operator_api, "get_by_dot_notation_list")
    def test_api_error_returns_none(self, mock_get_by_dot_notation_list):
        mock_get_by_dot_notation_list.side_effect = ApiError("mock_error")
        mock_value = "mock_value"
        mock_query = {
            "$or": [
                {"mock.dot.notation.1": mock_value},
                {"mock.dot.notation.1.#text": mock_value},
                {"mock.dot.notation.2": mock_value},
                {"mock.dot.notation.2.#text": mock_value},
            ]
        }

        self.assertIsNone(get_keywords_from_search_operator_query(mock_query))



""" Serializers used for the persistent query keyword REST API.
"""
from rest_framework_mongoengine.serializers import DocumentSerializer

from core_explore_keyword_app.components.persistent_query_keyword import (
    api as persistent_query_keyword_api,
)
from core_explore_keyword_app.components.persistent_query_keyword.models import (
    PersistentQueryKeyword,
)


class PersistentQueryKeywordSerializer(DocumentSerializer):
    """persistent query keyword serializer"""

    class Meta(object):
        model = PersistentQueryKeyword
        fields = ["id", "user_id", "content", "templates", "name"]
        read_only_fields = ("id", "user_id")

    def create(self, validated_data):
        """Create and return a new `PersistentQueryKeyword` instance, given the validated data."""
        # Create instance from the validated data and insert it in DB
        persistent_query_keyword = PersistentQueryKeyword(
            user_id=str(self.context["request"].user.id),
            content=validated_data["content"] if "content" in validated_data else None,
            templates=validated_data["templates"]
            if "templates" in validated_data
            else None,
            name=validated_data["name"] if "name" in validated_data else None,
        )
        return persistent_query_keyword_api.upsert(
            persistent_query_keyword, self.context["request"].user
        )

    def update(self, persistent_query_keyword, validated_data):
        """Update and return an existing `PersistentQueryKeyword` instance, given the validated
        data.
        """
        persistent_query_keyword.content = validated_data.get(
            "content", persistent_query_keyword.content
        )
        persistent_query_keyword.templates = validated_data.get(
            "templates", persistent_query_keyword.templates
        )
        persistent_query_keyword.name = validated_data.get(
            "name", persistent_query_keyword.name
        )
        return persistent_query_keyword_api.upsert(
            persistent_query_keyword, self.context["request"].user
        )


class PersistentQueryKeywordAdminSerializer(DocumentSerializer):
    """PersistentQueryKeyword Serializer"""

    class Meta(object):
        """Meta"""

        model = PersistentQueryKeyword
        fields = ["id", "user_id", "content", "templates", "name"]

    def create(self, validated_data):
        """
        Create and return a new `PersistentQueryKeyword` instance, given the validated data.
        """
        # Create data
        persistent_query_keyword = PersistentQueryKeyword(
            user_id=validated_data["user_id"],
            content=validated_data["content"] if "content" in validated_data else None,
            templates=validated_data["templates"]
            if "templates" in validated_data
            else None,
            name=validated_data["name"] if "name" in validated_data else None,
        )
        return persistent_query_keyword_api.upsert(
            persistent_query_keyword, self.context["request"].user
        )
